/*  Student Name: Wade Tregaskis
    Student Number: 02557793
    Estimated time spent on this Stage 1 assignment: Too much!
            More accurately, I guess somewhere between 60 and 90 hours.  But as
            always much of that time was just experimenting with C++, so, actual
            productive hours is probably more like 40, if that.

    Notes:  Project created by creating a copy of the IPC assignment, and using
      that as a starting point.  The reason for this is that said assignment
      already contained lots of code for linked lists, save/load of data etc,
      as well as many utility functions (e.g. caseInsensitiveStringCompare).
            Also note that due - ultimately - to some poor design choices, there
      is a lot of code just for handling the two "modes" of operation - my way
      of doing things, and the exact spec way.  If either 'mode' were cut
      there would be huge savings.  But, I like my mode. */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "species.h"


#define SPECIES_C_VERSION 1.1


/* int main(int argc, const char **argv)
    Handles the main event loop.  Accepts parameters, including filename, from the
    shell, or gets them from the user if not present  Loads the inital database,
    runs the main logical loop including main menu, and stores data back to disk. */

int main(int argc, const char **argv) {
    char *fileName;
    int choice = '\0', fileArgIndex;
    SpeciesList *defaultList;
    bool useStrictFileFormats;
    
    // mtrace(); // Generates malloc trace's - oooh so useful :)

    if (switchIsInArguments(argc, argv, 'v')) {
        cout << "Species version " << SPECIES_C_VERSION << " (base version " << SPECIES_H_VERSION << ")\n";
        cout << "   Written by Wade Tregaskis.  Free to distribute.\n";
        cout << "   Produced under hypothetical license for the Terran Central Exploration Bureau.\n";

        return 0;
    }
    
    if (switchIsInArguments(argc, argv, 'h')) {
        cout << "Usage:       " << argv[0] << " [-hvs] [file]\n";
        cout << "Parameters:\n     -h      Display this help\n"
                          << "     -v      Display version number [and quit]\n"
                          << "     -s      Use strict file format rules for loading and saving data\n"
                          << "     file    Data file containing species information\n";
        return 0;
    }

    useStrictFileFormats = switchIsInArguments(argc, argv, 's');

    if (useStrictFileFormats) {
        cout << "Using strict file formats.\n";
    }
    
    fileArgIndex = firstDataArgument(argc, argv);
    
    if (fileArgIndex > 0) {
        fileName = copyOfString(argv[fileArgIndex]);
    } else {
        cout << "Name of database from which to load species data (default is 'fact.dat'):  ";
        fileName = new char[FILENAME_BUFFER_SIZE];
        safeGetline(cin, fileName, FILENAME_BUFFER_SIZE);

        if (strlen(fileName) == 0) {
            delete[] fileName;
            fileName = copyOfString("fact.dat");
        }
    }
    
    defaultList = newListFromFile(fileName, useStrictFileFormats);
    
    if (!defaultList) {
        cout << "\nError: Unable to read species from \"" << fileName << "\" - "
             << "it may be missing or in use.\nIf you continue, a new [empty] "
             << "database will be created.\n\nDo you wish to continue?  ";

        if (!getBoolean(cin)) {
            return 0;
        }

        defaultList = newList(fileName);
    } else if (defaultList->numberOfItems == 0) {
        cout << "\nWarning: \"" << fileName << "\" contains no entries.\n";
    }

    while (1) {
        cout << endl;
        
        switch (choice = uiIndexedMenu("TCEB Species Classification Program", 8, ((defaultList->numberOfItems > 0) ? "Browse Species List" : NULL), 1, ((defaultList->numberOfItems > 0) ? "Search for Existing Species" : NULL), 2, "Add New Species", 3, NULL, 0, "Strict File Formatting", 6, ((defaultList->numberOfDataErrors > 0) ? "Data Errors" : NULL), 4, NULL, 0, "Exit Program", 5)) {
        case 1:
            uiDisplayList(defaultList); break;
        case 2:
            uiSearchList(defaultList); break;
        case 3:
            uiAddSpeciesToList(defaultList); break;
        case 4:
            uiExplainDataErrors(defaultList); break;
        case 5:
            if (saveListAs(defaultList, useStrictFileFormats)) {
                deleteList(defaultList);
                return 0;
            } else {
                break;
            }
        case 6:
            uiExplainStrictFileFormat(); break;
        }
    }
}

/* uiDisplayList(SpeciesList *list, bool format)
     Presents a list of all entries in 'list' to the user.  Note: It does not
     sort the database or interpret the results in any other way than needed to
     display them.
       Automatically asks the user for long or short form of list. */

void uiDisplayList(SpeciesList *list) {
    char choice = '\0';

    if (list->numberOfItems > 0) {
        cout << "Long or short list format (l/s)? ";
        choice = getChar(cin);
    }

    cout << endl;
    
    uiDisplayList(list, (toupper(choice) == 'L'));
}

/* uiDisplayList(SpeciesList *list, bool format)
     Presents a list of all entries in 'list' to the user.  Note: It does not
     sort the database or interpret the results in any other way than needed to
     display them.
       If 'format' is true, entries are displayed in long form, otherwise only
     in brief. */

void uiDisplayList(SpeciesList *list, bool format) {
    SpeciesListItem *current;
    
    if (!(list->numberOfItems)) {
        if (list->openedFrom) {
            cout << "There are no entries in the list \"" << list->openedFrom << "\".\n";
        } else {
            cout << "There are no entries to list.\n";
        }
    } else {
        current = list->first;

        if (list->openedFrom) {
            cout << "The list from \"" << list->openedFrom << "\" contains the following species:\n\n";
        }

        if (format) {
            outputFixedLength(cout, "NAME", 10, true);
            outputFixedLength(cout, "<PROPERTIES>", 30, true);
        } else {
            outputFixedLength(cout, "NAME", 40, true);
        }
        
        cout << "LIVING | FORM | TYPE\n\n";
        
        while (current) {
            outputItem(cout, current->target, format);
            current = current->next;
        }

        if (list->numberOfDataErrors) {
            cout << "\nFor information about item(s) marked with (#n), see 'Data Errors' in the main menu.\n";
        }
    }
}

/* uiAddSpeciesToList(SpeciesList *list)
     Inputs the title, artist and length of a new SpeciesItem, appending it to
     'list'. */

void uiAddSpeciesToList(SpeciesList *list) {
    char input[LINE_BUFFER_SIZE];
    SpeciesItem *result = NULL;
    PropertyItem *curProp = NULL, *lastProp = NULL;
    
    cout << "Adding new species to \"" << list->openedFrom << "\".\n\nSpecies name: ";
    safeGetline(cin, input, LINE_BUFFER_SIZE);

    if (strlen(input) == 0) {
        cout << "Treating lack of input as cancel - returning to menu.\n";
        return;
    }
    
    result = newSpecies(copyOfString(input));

    switch (uiIndexedMenu("Status:", 4, "Living", 1, "Extinct", 2, NULL, 0, "Unknown/None of the above", 3)) {
    case 1:
        result->living = LIVING_YES; break;
    case 2:
        result->living = LIVING_NO; break;
    case 3:
        result->living = UNKNOWN; break;
    }

    switch (uiIndexedMenu("Type:", 9, "Land Animal", 1, "Water Animal", 2, "Air Animal", 3, "Amphibian", 4, "Tree", 5, "Flower", 6, "Shrub", 7, NULL, 0, "Unknown/None of the above", 8)) {
    case 1:
        result->form = ANIMAL;
        result->type = ANIMAL_LAND;
        break;
    case 2:
        result->form = ANIMAL;
        result->type = ANIMAL_WATER;
        break;
    case 3:
        result->form = ANIMAL;
        result->type = ANIMAL_AIR;
        break;
    case 4:
        result->form = ANIMAL;
        result->type = ANIMAL_AMPHIBIAN;
        break;
    case 5:
        result->form = PLANT;
        result->type = PLANT_TREE;
        break;
    case 6:
        result->form = PLANT;
        result->type = PLANT_FLOWER;
        break;
    case 7:
        result->form = PLANT;
        result->type = PLANT_SHRUB;
        break;
    case 8:
        result->form = UNKNOWN;
        result->type = UNKNOWN;
        break;
    }

    cout << "Additional Properties (Enter a blank line when there are no more to add)\n";

    while (1) {
        cout << "Property Name: ";
        safeGetline(cin, input, LINE_BUFFER_SIZE);

        if (strlen(input)) {
            lastProp = curProp;

            curProp = newProperty(copyOfString(input), NULL);
            
            cout << "        Value: ";
            safeGetline(cin, input, LINE_BUFFER_SIZE);

            if (strlen(input)) {
                curProp->property = copyOfString(input);

                if (lastProp) {
                    lastProp->next = curProp;
                } else {
                    result->firstProperty = curProp;
                }
            } else {
                break;
            }
        } else {
            break;
        }
    }

    addSpeciesToList(list, result);

    cout << "\nNew species \"" << result->name << "\" added to \"" << list->openedFrom << "\".\n";
}

/* uiSearchList(SpeciesList *list)
     Uses repeatitive user input to search for a particular species, showing an
     increasingly narrow list of options until the final singular item is found,
     or no match is found, in which case the option to add a new entry is made
     available.
       Dev Note:  This code is bad.  It just screams "I am evil!  Recode me!".
     Unfortunately this is not easily possible with the current class-less dev
     model. */

void uiSearchList(SpeciesList *list) {
    const int LIVING_COUNT = 2, TYPE_COUNT = 7;
    char *qLiving[LIVING_COUNT], *qType[TYPE_COUNT];
    int i, living = UNKNOWN, form = UNKNOWN, type = UNKNOWN;
    int mLiving[LIVING_COUNT] = {LIVING_YES, LIVING_NO}, mType[TYPE_COUNT] = {ANIMAL_LAND, ANIMAL_WATER, ANIMAL_AIR, ANIMAL_AMPHIBIAN, PLANT_TREE, PLANT_FLOWER, PLANT_SHRUB};
    bool skipLiving, skipType;
    SpeciesList *current = NULL, *temp = NULL;
    char *holder, *holder2;

    if (!list) {
        cout << "\nInternal error - no list to search. (#2; list == NULL)\n";
        return;
    }

    if (list->numberOfItems == 0) {
        cout << "\nThere are no entries in \"" << list->openedFrom << "\".\n";
        return;
    } else if (list->numberOfItems == 1) { // Dude, c'mon...
        cout << endl;
        uiDisplayList(list, true);
        return;
    }

    cout << "Present a list of possible choices [vs asking for values] where possible?  ";
    
    if (!getBoolean(cin)) {
        uiSearchList_askQuestions(list, NULL);
        return;
    }

    cout << endl;

    // Well, there's a few ways of doing this...
    // ...what follows is not one of the pretty ones.

    for (i = 0; i < LIVING_COUNT; ++i) {
        qLiving[i] = ((findByLiving(list, mLiving[i])) ? livingToString(mLiving[i], false) : NULL);
    }

    skipLiving = isArrayVoid((void**)qLiving, 2);

    if (skipLiving) {
        current = list;
    } else {
        // The following line is bad - it assumes LIVING_COUNT == 2
        living = uiIndexedMenu("Status:", 4, qLiving[0], mLiving[0], qLiving[1], mLiving[1], NULL, 0, "Unknown/None of the Above", UNKNOWN);

        current = findAllByLiving(list, living);

        if (current->numberOfItems != list->numberOfItems) {
            uiSearchList_showUpdate(current);
        }
    }

    deleteStringArray(qLiving, LIVING_COUNT); // Prevent mega fat leaks

    if (uiSearchList_canMatch(list, current, living, form, type)) {
        if (current != list) {
            deleteList(current);
        }
        
        return;
    }

    for (i = 0; i < TYPE_COUNT; ++i) {
        qType[i] = ((findByType(current, mType[i])) ? typeToString(mType[i], false) : NULL);
    }

    skipType = isArrayVoid((void**)qType, 7);

    if (!skipType) {
        // The following line is bad - it assumes TYPE_COUNT == 7
        type = uiIndexedMenu("Type:", 9, qType[0], mType[0], qType[1], mType[1], qType[2], mType[2], qType[3], mType[3], qType[4], mType[4], qType[5], mType[5], qType[6], mType[6], NULL, 0, "Unknown/None of the Above", UNKNOWN);

        temp = current;
        current = findAllByType(temp, type);

        if (current->numberOfItems != temp->numberOfItems) {
            uiSearchList_showUpdate(current);
        }
        
        if (temp != list) {
            deleteList(temp);
        }
    }

    deleteStringArray(qType, TYPE_COUNT); // Prevent mega fat leaks
    
    if (uiSearchList_canMatch(list, current, living, form, type)) {
        if (current != list) {
            deleteList(current);
        }
        
        return;
    }

    if (type != UNKNOWN) {
        form = deriveFormFromType(type);
        temp = findAllByForm(current, form);

        if (temp->numberOfItems > 0) {
            holder = formToString(form, true);
            holder2 = typeToString(type, true);
            cout << "Is it safe to assume this " << holder2 << " is a " << holder << "?  ";
            delete[] holder;
            delete[] holder2;

            if (!getBoolean(cin)) {
                form = UNKNOWN;
                deleteList(temp);
            } else {
                if (current->numberOfItems != temp->numberOfItems) {
                    uiSearchList_showUpdate(temp);
                }

                if (current != list) {
                    deleteList(current);
                }

                current = temp;
                
                if (uiSearchList_canMatch(list, current, living, form, type)) {
                    deleteList(current);
                    return;
                }
            }
        } else {
            form = UNKNOWN;
            deleteList(temp);
        }
    }

    uiSearchList_askQuestions(current, newQuestionList(living, form, type, true));

    if (current != list) {
        deleteList(current);
    }
}

/* uiSearchList_askQuestions(SpeciesList *list, QuestionItem *startList)
     Asks the user questions based on the properties of the items in 'list', such
     that a match (or lack of) can be made.  If 'current' is not NULL, all questions
     in 'current' are treated as having already been asked, and so are not asked
     again.
       Very important note: 'list' is not changed in any way, but 'startList', if
     it is not NULL, _will_ be changed.  By changed, any questions asked of the
     user by this function will be appended to the end of 'startList'. */

void uiSearchList_askQuestions(SpeciesList *list, QuestionItem *questionList) {
    QuestionItem *firstQuestion = NULL, *currentQuestion = NULL, *lastQuestion = NULL;
    SpeciesList *currentList = NULL;
    char *input = NULL;
    int lastCount;
    bool isRelevant;

    firstQuestion = questionList;
    currentQuestion = questionList;

    while (currentQuestion) {
        lastQuestion = currentQuestion;
        currentQuestion = currentQuestion->next;
    }

    if (list == NULL) {
        deleteQuestionList(firstQuestion);
        return;
    }

    currentList = copyOfList(list);
    lastCount = currentList->numberOfItems;
    
    while ((currentQuestion = generateNextQuestion(currentList, firstQuestion, true)) != NULL) {
        isRelevant = questionIsRelevant(currentList, currentQuestion);
        
        if (isRelevant) {
            if (currentQuestion->isBoolean) {
                cout << "Is the species " << currentQuestion->question << "?  ";
            } else {
                cout << "Enter a value below, or 'remove' to ignore all species with this property.\n" << currentQuestion->question << ":  ";
            }

            input = new char[LINE_BUFFER_SIZE];
            safeGetline(cin, input, LINE_BUFFER_SIZE);
            currentQuestion->answer = copyOfString(input);
            delete[] input;

            if (caseInsensitiveStringCompare(currentQuestion->answer, "remove") == 0 || ((currentQuestion->isBoolean) && (!boolFromString(currentQuestion->answer)))) {
                removeSpeciesWithProperty(currentList, currentQuestion->question);
            } else {
                removeSpeciesWithDifferentProperty(currentList, currentQuestion->question, ((currentQuestion->isBoolean) ? NULL : currentQuestion->answer));
            }
        }
        
        if (lastQuestion != NULL) {
            lastQuestion->next = currentQuestion;
            lastQuestion = currentQuestion;
        } else {
            firstQuestion = currentQuestion;
            lastQuestion = currentQuestion;
        }

        if (isRelevant) {
            if (currentList->numberOfItems != lastCount) {
                uiSearchList_showUpdate(currentList);

                if (uiSearchList_canMatch(list, currentList, firstQuestion)) {
                    deleteQuestionList(firstQuestion);
                    deleteList(currentList);

                    return;
                }
            } else {
                cout << "That didn't help.  Hmm...probably an internal error...#1.\n";
            }
        }

        lastCount = currentList->numberOfItems;
    }

    cout << "No more questions to ask!  All remaining items appear to be identical.\n";

    if (!uiSearchList_askIfCorrect()) {
        uiSearchList_askToAdd(list, firstQuestion);
    }
    
    deleteQuestionList(firstQuestion);
    deleteList(currentList);
}

/* uiSearchList_showUpdate(SpeciesList *list)
     Support function for uiSearchList.  Not recommended for use outside this.
     Outputs the number of items in 'list', and if they are less than
     BRIEF_LIST_THRESHOLD displays them.  Otherwise simply notes 'Too many to
     show'. */

void uiSearchList_showUpdate(SpeciesList *list) {
    cout << "This reduces the list of possible matches to " << list->numberOfItems << " species.  ";

    if (list->numberOfItems > 0) {
        if (list->numberOfItems <= BRIEF_LIST_THRESHOLD) {
            cout << "Listed below.\n\n";
            uiDisplayList(list, true);
            cout << endl;
        } else {
            cout << "Too many to show.\n\n";
        }
    }
}

/* uiSearchList_askIfCorrect(void)
     Support function for uiSearchList.  Not recommended for use outside this.
     Very simply asks the user if the current match is what they were looking for,
     and returns true if they enter 'y' or 'Y'.  False otherwise. */

bool uiSearchList_askIfCorrect(void) {
    cout << "Is this the species you were looking for?  ";
    return (getBoolean(cin));
}

/* uiSearchList_canMatch(SpeciesList *original, SpeciesList *sublist, int living, int form, int type)
     Support function for uiSearchList.  Not recommended for use outside this.
     Determines if the search can be finished now by virtue of the fact that if
     there is 0 or 1 item in 'sublist', a decision must be made now.  Offers the
     users the opportunity to add a new entry to 'original' if there is no match.
     Returns 'true' if there was a match, 'false' if not (e.g. because there is
     more than one entry in 'sublist') */

bool uiSearchList_canMatch(SpeciesList *original, SpeciesList *sublist, int living, int form, int type) {
    if (sublist->numberOfItems < 1) {
        cout << "No match found.\n\n";
        uiSearchList_askToAdd(original, living, form, type);
        return true;
    } else if (sublist->numberOfItems == 1) {
        if (!uiSearchList_askIfCorrect()) {
            uiSearchList_askToAdd(original, living, form, type);
        }

        return true;
    } else {
        return false;
    }
}

/* uiSearchList_canMatch(SpeciesList *original, SpeciesList *sublist, QuestionItem *questions)
     Support function for uiSearchList.  Not recommended for use outside this.
     Determines if the search can be finished now by virtue of the fact that if
     there is 0 or 1 item in 'sublist', a decision must be made now.  Offers the
     users the opportunity to add a new entry to 'original' if there is no match.
     Returns 'true' if there was a match, 'false' if not (e.g. because there is
     more than one entry in 'sublist') */

bool uiSearchList_canMatch(SpeciesList *original, SpeciesList *sublist, QuestionItem *questions) {
    if (sublist->numberOfItems < 1) {
        cout << "No match found.\n\n";
        uiSearchList_askToAdd(original, questions);
        return true;
    } else if (sublist->numberOfItems == 1) {
        if (!uiSearchList_askIfCorrect()) {
            uiSearchList_askToAdd(original, questions);
        }

        return true;
    } else {
        return false;
    }
}

/* uiSearchList_askToAdd(SpeciesList *list, int living, int form, int type)
     Support function for uiSearchList.  Not recommended for use outside this.
     Asks the user if they would like to add a new species, and does so if they
     direct so.  Otherwise does nothing. */

void uiSearchList_askToAdd(SpeciesList *list, int living, int form, int type) {
    cout << "Would you like to add a new species?  ";

    if (getBoolean(cin)) {
        cout << "Well too bad, you can't - I haven't coded this bit yet. :P\n";
        // FLAG - Do code to add new species here
    }
}

/* uiSearchList_askToAdd(SpeciesList *list, QuestionItem *questions)
     Support function for uiSearchList.  Not recommended for use outside this.
     Asks the user if they would like to add a new species, and does so if they
     direct so.  Otherwise does nothing. */

void uiSearchList_askToAdd(SpeciesList *list, QuestionItem *questions) {
    cout << "Would you like to add a new species?  ";

    if (getBoolean(cin)) {
        cout << "Well too bad, you can't - I haven't coded this bit yet. :P\n";
        // FLAG - Do code to add new species here
    }
}	  
	  
/* uiIndexedMenu(const char *title, int optionCount, ...)
     Draws a generic menu with the title of 'title', and containing 'optionCount'
     number of items.  After this, each option is passed as a pair of arguments -
     first a const char* to the option name.  Then an integer which is returned if
     that option is choosen.
       To display a blank link in the menu as a whitespace separator, pass NULL
     as the item name and 0 as the return value.  If you pass NULL but a return
     value of anything else, that entry will be ignored. */

int uiIndexedMenu(const char *title, int optionCount, ...) {
    va_list arg;
    int *menuList, realCount = 0;
    const char *name = NULL;
    char choice = '\0';

    if (optionCount > 0) {
        menuList = new int[optionCount];
        
        cout << "--<== " << title << " ==>--\n\n";

        va_start(arg, optionCount);

        for(int i = 0; i < optionCount; ++i) {
            name = va_arg(arg, const char*);
            menuList[realCount] = va_arg(arg, int);

            if (name) {
                cout << "  " << realCount + 1 << ". " << name << endl;
                ++realCount;
            } else {
                if (menuList[realCount] == 0) {
                    cout << endl;
                }
            }
        }

        va_end(arg);

        cout << endl;

        while (1) {
            cout << "Selection: ";
            choice = getChar(cin);
            
            if ((choice < '1') || ((int)choice > (realCount + '0'))) {
                cout << "Invalid selection - please enter a number between 1 and " << realCount << ".\n";
            } else {
                break;
            }
        }

        realCount = menuList[choice - '1']; // Note that realCount is used here only as a temporary variable
        delete[] menuList;
        
        return realCount;
    } else {
        return 0;
    }
}

/* uiExplainDataErrors(SpeciesList *list)
     Gives the user detailed information on what data errors the program can
     identify, why they may have occured, and how to fix them.  If 'list' is not
     NULL it identifies all entries in it which have data errors, and attempts to
     explain them. */

void uiExplainDataErrors(SpeciesList *list) {
    SpeciesListItem *currentItem;
    SpeciesItem *current;
    
    cout << "\nA data error on any entry indicates that there is something wrong with that entry.\nFor instance, it may be a Land Animal, and yet be marked as a Plant.  Or there may\nhave been some problem with reading it in from the source file.  Alternatively it may\nappear to have been corrupted in memory.  Or some other problem which gives the end\nresult that something in the entry just isn't right.\n\n";
    cout << "You can identify affected entries in any list.  At the end of the line - after their\nname, living, form and type properties - a #n will be present, in braces, where n is\nsome error number.  You may also find that if the error is a simple one the list\nentry will include the suggested fix.\n\n";

    if (list->numberOfDataErrors > 0) {
        if (list->numberOfDataErrors > 1) {
            cout << "In the list \"" << list->openedFrom << "\", the entries marked as in error are:\n\n";
        } else {
            cout << "In the list \"" << list->openedFrom << "\", the entry marked as in error is:\n\n";
        }

        currentItem = list->first;

        while (currentItem) {
            current = currentItem->target;

            if (current->dataError != 0) {
                outputItem(cout, current, true);
            }

            currentItem = currentItem->next;
        }

        cout << "\nMatch the error numbers with their meaning below, and try to fix the problem.\n\n";
    } else {
        cout << "Meanings of Error Numbers:\n\n";
    }

    cout << " #1   Form is not appropriate for specified type.  e.g. 'Land Animal' for a 'Plant'.\n";
    cout << " #2   Amphibian type is not allowed in strict formatting mode.\n";
    cout << " #3   Implicit declaration of not-living not allowed in strict formatting mode.\n";
    cout << " #4   More than " << STRICT_PROPERTY_COUNT_LIMIT << " properties not allowed in strict formatting mode.\n";
    cout << " #5   Properties may not have values of 'remove' - this is a reserved keyword.\n";
}

/* uiExplainStrictFileFormat(void)
     Provides the user with details on what it means to enforce strict file
     formats, and how they can use one or the other. */

void uiExplainStrictFileFormat(void) {
    cout << "\nSpecies uses a standard file format to store data, as defined by the TCEB.  However,\nit is capable of extending this file format in several ways in order to include more\ninformation and data.\n\n";
    cout << "This can cause problems when sharing data files with other programs, which may not\nunderstand this extra information.  As such, it is possible to make Species enforce\nstrict file formatting, which disallows use of any special extensions.  To do this,\npass the '-s' flag when running Species.\n\n";
    cout << "Please note that if you try to use an extended format file in strict formatting mode,\nthe extra's [if present] of the extended format will be marked as errors.\n";
}

/* displayRepeat(ostream& out, char ch, int times)
     Outputs to the stream 'out' a sequence of identical characters - 'char ch'.
     It does so to the tune of 'times'.
       Note: The use of the 'times' parameter instead of a local variable is not
     erroneous. */

void displayRepeat(ostream& out, char ch, int times) {
    while (times > 0) {
        out << ch;
        --times;
    }
}

/* removeOutsideWhitespace(char **data)
Takes the 'data' and removes any whitespace before or after any text in it.
Note that 'data' is a pointer to a pointer to a string (phew!), so for a
normal *char pointer, you'll need to prefix & to the pointer name when
passing to this function. */

void removeOutsideWhitespace(char **data) {
    int textStart, textEnd = 0, i, length, newLength;
    char *result, *realData;

    realData = *data;
    length = strlen(realData);

    textStart = length;

    for (i = 0; i < length; ++i) {
        if (!isspace(realData[i])) {
            textStart = i;
            break;
        }
    }

    for (i = length; i > 0; --i) {
        if (!isspace(realData[i - 1])) {
            textEnd = i;
            break;
        }
    }

    newLength = (textEnd - textStart);

    if (newLength > 0) { // malloc doesn't like negative values
        result = new char[newLength + 1];

        for (i = 0; i < newLength; ++i) {
            result[i] = realData[textStart + i];
        }

        result[newLength] = '\0';
    } else { // Indicates the original string was pure whitespace
        result = new char[1];
        result[0] = '\0';
    }

    delete[] realData;
    *data = result;
}

/* outputFixedLength(ostream& out, const char *data, int maxLength, bool left)
     Outputs the cstring 'data' to 'out', establishing an output of fixed length,
     set by 'maxLength', and aligned to left or right, depending on 'left'.  Most
     useful for the fact that if 'data' is too long it will be truncated and an
     elipse automagically appended to indicate this. */

void outputFixedLength(ostream& out, const char *data, int maxLength,
                       bool left) {
    char *result;
    int i;

    if (left) {
        cout.setf(ios::left);
    } else {
        cout.setf(ios::right);
    }

    cout.width(maxLength);

    if (strlen(data) > (unsigned int)maxLength) { // Need to truncate it
        result = new char[maxLength + 1];
        strncpy(result, data, maxLength - 3);

        for (i = maxLength - 3; i < maxLength; ++i) {
            result[i] = '.';
        }

        result[maxLength] = '\0';
        
        cout << result;

        delete[] result;
    } else { // Fine as it is
        cout << data;
    }
}

/* outputItem(ostream& out, SpeciesItem *item, bool format)
     Prints to the stream 'out' the contents of 'item', in human readable form.
     If 'format' is true, properties are displayed as well, each on a new line
     and indented.  If 'format' is false, only the name, living status, form and
     type are shown, all on one line. */

void outputItem(ostream& out, SpeciesItem *item, bool format) {
    PropertyItem *curProperty = NULL;
    char *holder;
    
    outputFixedLength(out, item->name, 42, true);

    switch (item->living) {
    case LIVING_NO:
        outputFixedLength(out, "No", 7, true); break;
    case LIVING_YES:
        outputFixedLength(out, "Yes", 7, true); break;
    default:
        outputFixedLength(out, " ?", 7, true); break;
    }

    switch (item->form) {
    case ANIMAL:
        outputFixedLength(out, "Anml", 7, true); break;
    case PLANT:
        outputFixedLength(out, "Plnt", 7, true); break;
    default:
        outputFixedLength(out, " ?", 7, true);
    }

    switch (item->type) {
    case ANIMAL_LAND:
        out << "Land"; break;
    case ANIMAL_WATER:
        out << "Water"; break;
    case ANIMAL_AIR:
        out << "Air"; break;
    case ANIMAL_AMPHIBIAN:
        out << "Amph"; break;
    case PLANT_TREE:
        out << "Tree"; break;
    case PLANT_FLOWER:
        out << "Flower"; break;
    case PLANT_SHRUB:
        out << "Shrub"; break;
    default:
        out << " ?"; break;
    }

    if ((item->form == UNKNOWN) && (item->type != UNKNOWN)) {
        holder = formToString(deriveFormFromType(item->type), true);
        out << " [" << holder << "?]";
        delete[] holder;
    }

    if (item->dataError != 0) {
        out << " (#" << item->dataError << ")";
    }
    
    if (format) {
        curProperty = item->firstUnknown;

        while (curProperty) {
            out << endl;
            displayRepeat(out, ' ', 10);
            out << curProperty->name;

            curProperty = curProperty->next;
        }
        
        curProperty = item->firstProperty;

        while (curProperty) {
            out << endl;
            displayRepeat(out, ' ', 10);
            out << curProperty->name << ":  " << curProperty->property;

            curProperty = curProperty->next;
        }
    }

    out << endl;
}

/* newList(char *sourceFile) {
     Creates a new SpeciesItem list, designated as being from the file
     'sourceFile', and returns a pointer to it. */

SpeciesList* newList(char *sourceFile) {
    SpeciesList *result;

    result = new SpeciesList;
    result->first = NULL;
    result->last = NULL;
    result->openedFrom = sourceFile;
    result->numberOfItems = 0;
    result->numberOfDataErrors = 0;

    return result;
}

/* copyOfList(SpeciesList *list)
     Returns a copy of 'list'.  This means a completely new instance of
     SpeciesList, with completely unique instances of SpeciesListItem's in it.
     These however do still point to the same SpeciesItem's as the original list.
     The 'original' property of each SpeciesListItem will be NULL, as if it were
     created 'normally'. */

SpeciesList* copyOfList(SpeciesList *list) {
    SpeciesList *result = NULL;
    SpeciesListItem *current = NULL;

    if (list) {
        result = newList(NULL);

        current = list->first;

        while (current) {
            addSpeciesToList(result, current->target);
            current = current->next;
        }
    }

    return result;
}

/* newListFromFile(char *fileName, bool strict)
     Reads from the file 'fileName' data about species.  Turns this into a
     SpeciesList, and returns this.  If 'strict' is true, strict formatting is
     used and any extensions are marked as errors.  Otherwise only true errors
     are marked as errors. */

SpeciesList* newListFromFile(char *fileName, bool strict) {
    SpeciesList *result = NULL;
    bool newEntry = true;
    ifstream fin;
    char curLine[LINE_BUFFER_SIZE], *propName = NULL, *propData = NULL;
    SpeciesItem *tempItem = NULL;
    int propValue, propType, propCount = 0;
    PropertyItem *tempProp = NULL, *lastProp = NULL, *tempUnknown = NULL, *lastUnknown = NULL;

    if (fileName == NULL) {
        return NULL;
    }
    
    fin.open(fileName);

    if (fin.fail()) {
        return NULL; // unable to open file
    }

    result = newList(fileName);
    
    while(!fin.eof()) {
        safeGetline(fin, curLine, LINE_BUFFER_SIZE);

        if (strlen(curLine) == 0) { // Empty line
            if (!newEntry) {
                if ((tempItem->type != UNKNOWN) && (deriveFormFromType(tempItem->type) != tempItem->form)) {
                    ++(result->numberOfDataErrors);
                    tempItem->dataError = 1; // Type not valid for given form
                }

                if (strict && (tempItem->type == ANIMAL_AMPHIBIAN)) {
                    ++(result->numberOfDataErrors);
                    tempItem->dataError = 2; // Amphibian type not supported [strict]
                }

                if (strict && (tempItem->living == LIVING_NO)) {
                    ++(result->numberOfDataErrors);
                    tempItem->dataError = 3; // Implicit spec of not living not supported [strict]
                }

                if (strict && (propCount > STRICT_PROPERTY_COUNT_LIMIT)) {
                    ++(result->numberOfDataErrors);
                    tempItem->dataError = 4; // Too many properties [strict]
                }
                
                addSpeciesToList(result, tempItem);
                
                lastProp = NULL;
                tempProp = NULL;
                tempUnknown = NULL;
                lastUnknown = NULL;
                propCount = 0;
            }

            newEntry = true;
        } else {
            if (newEntry) {
                tempItem = newSpecies(copyOfString(curLine));
                
                newEntry = false;
            } else {
                ++propCount;
                
                if (curLine[0] == '#') { // Defined property
                    propName = propertyNameFromLine(curLine);
                    propData = propertyValueFromLine(curLine);

                    if (caseInsensitiveStringCompare(propData, "remove") == 0) {
                        ++(result->numberOfDataErrors);
                        tempItem->dataError = 5; // 'remove' not a valid proerty value
                    }
                    
                    lastProp = tempProp;

                    tempProp = newProperty(propName, propData);

                    if (lastProp) {
                        lastProp->next = tempProp;
                    } else {
                        tempItem->firstProperty = tempProp;
                    }
                } else { // Hopefully a known property
                    propType = determinePropertyFromString(curLine, propValue);

                    switch (propType) {
                    case PROP_FORM:
                        tempItem->form = propValue; break;
                    case PROP_TYPE:
                        tempItem->type = propValue; break;
                    case PROP_LIVING:
                        tempItem->living = propValue; break;
                    default: // unknown
                        lastUnknown = tempUnknown;

                        tempUnknown = newProperty(copyOfString(curLine), NULL);

                        if (lastUnknown) {
                            lastUnknown->next = tempUnknown;
                        } else {
                            tempItem->firstUnknown = tempUnknown;
                        }
                    }
                }
            }
        }
    }

    fin.close();

    return result;
}

/* deleteList(SpeciesList *target)
     Deallocates all memory used by the SpeciesItem list and destroys it. */

void deleteList(SpeciesList *target) {
    SpeciesListItem *currentItem, *nextItem;
    
    if (target) {
        if (target->first) {
            currentItem = target->first;

            while (currentItem) {
                if (currentItem->next) {
                    nextItem = currentItem->next;
                } else {
                    nextItem = NULL;
                }

                deleteListItem(currentItem);
                currentItem = nextItem;
            }
        }

        if (target->openedFrom) {
            delete[] (target->openedFrom);
        }

        delete target;
    }
}

/* newSpecies(char *name)
     Returns a pointer to a newly allocated SpeciesItem struct, setting the
     SpeciesItem name to the initial value as passed in 'name'. */

SpeciesItem* newSpecies(char *name) {
    return newSpecies(name, UNKNOWN, UNKNOWN, UNKNOWN, NULL);
}

/* newSpecies(char *name, char *form, char *type, int living)
     Returns a pointer to a newly allocated SpeciesItem struct, setting the
     SpeciesItem name & initial properties as passed. */

SpeciesItem* newSpecies(char *name, char *form, char *type, int living) {
    return newSpecies(name, formFromString(form), typeFromString(type), living, NULL);
}

/* newSpecies(char *name, int form, int type, int living, PropertyItem *firstProp)
     Returns a pointer to a newly allocated SpeciesItem struct, setting the
     SpeciesItem name & initial properties as passed. */

SpeciesItem* newSpecies(char *name, int form, int type, int living, PropertyItem *firstProp) {
    SpeciesItem *temp;

    temp = new SpeciesItem;
    temp->name = name;
    temp->living = living;
    temp->form = form;
    temp->type = type;
    temp->dataError = 0;
    temp->firstProperty = firstProp;
    temp->firstUnknown = NULL;
    temp->claims = 0;

    return temp;
}

/* newProperty(char *name, char *value)
     Returns a new PropertyItem with the 'name' and 'value' passed. */

PropertyItem* newProperty(char *name, char *value) {
    PropertyItem *temp;
    
    temp = new PropertyItem;
    temp->name = name;
    temp->property = value;
    temp->next = NULL;

    return temp;
}

/* newQuestionItem(char *question)
     Creates a new QuestionItem instance and assigns it the value 'question'.
     Returns this instance. */

QuestionItem* newQuestionItem(char *question) {
    return newQuestionItem(question, NULL);
}

/* newQuestionItem(char *question, char *answer)
     Creates a new QuestionItem instance and assigns it the values 'question' and
     'answer'.  Returns this instance. */

QuestionItem* newQuestionItem(char *question, char *answer) {
    return newQuestionItem(question, answer, false);
}

/* newQuestionItem(char *question, char *answer, bool boolean)
     Creates a new QuestionItem instance and assigns it the values 'question',
     'answer' and 'boolean'.  Returns this instance. */

QuestionItem* newQuestionItem(char *question, char *answer, bool boolean) {
    QuestionItem *temp;

    temp = new QuestionItem;
    temp->next = NULL;
    temp->question = question;
    temp->answer = answer;
    temp->isBoolean = boolean;
    return temp;
}


/* newQuestionList(int living, int form, int type, bool ignoreUnknown)
     Creates a new list of questions for the initial so-called boolean properties
     'living', 'form' and 'type'.  If 'ignoreUnknown' is true, nodes will not be
     added if any of said properties == UNKNOWN, otherwise they will be added
     regardless. */

QuestionItem* newQuestionList(int living, int form, int type, bool ignoreUnknown) {
    QuestionItem *a = NULL, *b = NULL, *c = NULL;

    if (!ignoreUnknown || (living != UNKNOWN)) {
        a = newQuestionItem(livingToString(living, false), boolToString(true), true);
    }

    if (!ignoreUnknown || (form != UNKNOWN)) {
        if (a == NULL) {
            a = newQuestionItem(formToString(form, false), boolToString(true), true);
        } else {
            b = newQuestionItem(formToString(form, false), boolToString(true), true);
            a->next = b;
        }
    }

    if (!ignoreUnknown || (type != UNKNOWN)) {
        if (a == NULL) {
            a = newQuestionItem(typeToString(type, false), boolToString(true), true);
        } else if (b == NULL) {
            b = newQuestionItem(typeToString(type, false), boolToString(true), true);
            a->next = b;
        } else {
            c = newQuestionItem(typeToString(type, false), boolToString(true), true);
            b->next = c;
        }
    }

    return a;
}

/* deleteListItem(SpeciesListItem *item)
     Deletes the SpeciesListItem 'target'.  May or may not delete the SpeciesItem
     pointed to by the list item, depending on the list item's 'claim' property.
     'target' will always be NULL after this function. */

void deleteListItem(SpeciesListItem *item) { 
    if (item) {
        if (item->target) {
            --(item->target->claims);

            if (item->target->claims < 1) {
                deleteSpecies(item->target);
            }
        }

        delete item;
    }
}

/* deleteSpecies(SpeciesItem *target)
     Deletes the SpeciesItem 'target', meaning removes it and all its data from
     memory.  After calling this 'target' will no longer be a valid pointer and
     should not be used.  Note that this does not care what the value of the
     'claims' property is, so be very careful that you do care and check before
     you use it. */

void deleteSpecies(SpeciesItem *target) {
    if (target) {
        if (target->name) {
            delete[] (target->name);
        }

        deletePropertyList(target->firstProperty);
        deletePropertyList(target->firstUnknown);
                
        delete target;
    }
}

/* deletePropertyList(PropertyItem *list)
     Deletes all items in 'list'. */

void deletePropertyList(PropertyItem *list) {
    PropertyItem *next, *current = list;

    while (current) {
        next = current->next;
        deleteProperty(current);
        current = next;
    }
}

/* deleteProperty(PropertyItem *item)
     Deletes 'item', deallocating all memory associated with it.  Does not change
     any item it may or may not be pointing to with 'next'. */

void deleteProperty(PropertyItem *item) {
    if (item) {
        if (item->name) {
            delete[] item->name;
        }

        if (item->property) {
            delete[] item->property;
        }

        delete item;
    }
}

/* void deleteQuestionList(QuestionItem *list)
     Deletes 'list' and all further QuestionItem's it points to; i.e. if it's the
     start of a list, the entire list will be deleted.  Who would have thought. */

void deleteQuestionList(QuestionItem *list) {
    QuestionItem *next, *current = NULL;

	 if (list) {
	  current = list;
	  
    while (current) {
        next = current->next;
        deleteQuestion(current);
        current = next;
    }
	 }
}

/* deleteQuestion(QuestionItem *item)
     Deletes 'item', deallocating all memory associated with it.  Does not change
	  any item it may or may not be pointing to with 'next'. */

void deleteQuestion(QuestionItem *item) {
    if (item) {
        if (item->question) {
            delete[] item->question;
        }

        if (item->answer) {
            delete[] item->answer;
        }

        delete item;
    }
}

/* addSpeciesToList(SpeciesList *list, SpeciesItem *entry)
     Adds the SpeciesItem 'entry' to the SpeciesItem list 'list', performing all
     the voodoo required to make everything work properly. */

void addSpeciesToList(SpeciesList *list, SpeciesItem *entry) {
    SpeciesListItem *item;

    if (entry) {
        item = new SpeciesListItem;
        item->owner = list;
        item->target = entry;
        ++(entry->claims);
        
        if (list->last) {
            item->previous = list->last;
            list->last->next = item;
        } else {
            list->first = item;
            item->previous = NULL;
        }

        item->next = NULL;
        list->last = item;
        
        ++(list->numberOfItems);
    }
}

/* copySpeciesToList(SpeciesList *list, SpeciesListItem *entry)
     Copies the SpeciesItem contained in 'entry' to the SpeciesItemList 'list'.  The
     'entry' item will not be moved from it's present location. */

void copySpeciesToList(SpeciesList *list, SpeciesListItem *entry) {
    SpeciesListItem *item;

    if (entry && entry->target) {
        item = new SpeciesListItem;
        item->owner = list;

        if (entry->original) {
            item->original = entry->original;
        } else {
            item->original = entry;
        }
        
        item->target = entry->target;
        ++(entry->target->claims);

        if (list->last) {
            item->previous = list->last;
            list->last->next = item;
        } else {
            list->first = item;
            item->previous = NULL;
        }

        item->next = NULL;
        list->last = item;

        ++(list->numberOfItems);
    }
}

/* removeSpeciesFromList(SpeciesListItem *item) {
     Removes the SpeciesItem from the list, if it is in it.  As with
     addSpeciesToList(), this performs all the pointer manipulations required to
     ensure this is a seamless process.  Returns true if the SpeciesItem was in
     the list and was successfully removed, false otherwise.
       Note that this doesn't _delete_ the SpeciesItem list item or SpeciesItem. */

bool removeSpeciesFromList(SpeciesListItem *item) {
    if (item->owner == NULL) {
        return false;
    }
    
    if (item->next) {
        if (item->previous) { // Middle of the list
            item->next->previous = item->previous;
            item->previous->next = item->next;
        } else { // First in list
            item->owner->first = item->next;
            item->next->previous = NULL;
        }
    } else {
        if (item->previous) { // Last in list
            item->owner->last = item->previous;
            item->previous->next = NULL;
        } else { // Only item in list
            item->owner->first = NULL;
            item->owner->last = NULL;
        }
    }

    --(item->owner->numberOfItems);

    return true;
}

/* numberOfProperties(SpeciesItem *item)
     Returns the number of properties 'item' contains.  Very hacked. */

int numberOfProperties(SpeciesItem *item) {
    int count = 0;
    PropertyItem *cursor;

    if (item) {
        if (item->living != UNKNOWN) {
            ++count;
        }

        if (item->form != UNKNOWN) {
            ++count;
        }

        if (item->type != UNKNOWN) {
            ++count;
        }

        cursor = item->firstUnknown;

        while (cursor) {
            ++count;
            cursor = cursor->next;
        }

        cursor = item->firstProperty;

        while (cursor) {
            ++count;
            cursor = cursor->next;
        }
    }

    return count;
}

/* propertyNameByIndex(SpeciesItem *item, int n)
     Returns a copy of the property that corresponds to the mythical index 'n'.
     Note this function is very inefficient - the index is mapped across two
     linked lists and multiple int variables.  Try to avoid it if possible.
     Returns NULL if the index given is less than zero or greater than the number
     of properties. */

PropertyItem* propertyByIndex(SpeciesItem *item, int n) {
    PropertyItem *cursor;
    int count = -1;

    if (n < 0 || !item) {
        return NULL;
    }

    if (item->living != UNKNOWN) {
        ++count;

        if (count == n) {
            return newProperty(livingToString(item->living, false), NULL);
        }
    }

    if (item->form != UNKNOWN) {
        ++count;

        if (count == n) {
            return newProperty(formToString(item->form, false), NULL);
        }
    }

    if (item->type != UNKNOWN) {
        ++count;

        if (count == n) {
            return newProperty(typeToString(item->type, false), NULL);
        }
    }

    cursor = item->firstUnknown;

    while (cursor) {
        ++count;

        if (count == n) {
            return newProperty(copyOfString(cursor->name), NULL);
        }

        cursor = cursor->next;
    }

    cursor = item->firstProperty;

    while (cursor) {
        ++count;

        if (count == n) {
            return newProperty(copyOfString(cursor->name), copyOfString(cursor->property));
        }

        cursor = cursor->next;
    }

    return NULL;
}

/* generateNextQuestion(SpeciesList *list, QuestionItem *pastQuestions, bool tryToBeRandom)
     Returns a QuestionItem instance containing the next question to be asked of
     the species in 'list'.  If 'pastQuestions' is not NULL, it will ensure it
     does not return any question that is present in 'pastQuestions'.  If
     'tryToBeRandom' is true, this function will obviously try to pick the
     question somewhat randomly (i.e. such that the order is not predictable),
     but makes no guarantees.
       If there are no questions that can be asked of 'list' that don't already
     appear in 'pastQuestions', returns NULL. */

QuestionItem* generateNextQuestion(SpeciesList *list, QuestionItem *pastQuestions, bool tryToBeRandom) {
    SpeciesListItem *currentItem = NULL;
    SpeciesItem *current = NULL;
    PropertyItem *temp = NULL;
    QuestionItem *result = NULL, *fallback = NULL;
    int i;

    if (list) {
        currentItem = list->first;

        while (currentItem) {
            current = currentItem->target;

            for (i = 0; i < numberOfProperties(current); ++i) {
                temp = propertyByIndex(current, i);

                if (temp && !questionIsInList(temp->name, pastQuestions)) {
                    result = newQuestionItem(copyOfString(temp->name), NULL, (temp->property == NULL));

                    deleteProperty(temp);

                    if (fallback) {
                        deleteQuestion(fallback);
                    }
                    
                    if (!tryToBeRandom || ((int)(list->numberOfItems * (rand() / (RAND_MAX + 1.0)))) == 0) {
                        return result;
                    } else {
                        fallback = result;
                    }
                } else {
                    deleteProperty(temp);
                }
            }

            currentItem = currentItem->next;
        }
    }
	 
    return fallback;
}

/* questionIsInList(const char *question, QuestionItem *list)
     Returns true if the question 'question' appears in 'list'. */

bool questionIsInList(const char *question, QuestionItem *list) {
    QuestionItem *current = list;

    while (current) {
        if (caseInsensitiveStringCompare(question, current->question) == 0) {
            return true;
        }

        current = current->next;
    }

    return false;
}

/* isInList(SpeciesList *list, SpeciesItem *item, bool ignoreName)
     Returns 'true' if 'item' appears in 'list', by comparing not the physical
     addresses (see exactIsInList), but the properties themselves, using string
     comparisons and so forth.  If 'ignoreName' is true, the names of each
     species will be ignored - useful for finding duplicate entries under
     different names.  Returns false if no match could be made.
       Note that 'item' is considered matching anything in 'list' if it has all
     the same properties as any item in 'list'.  If said item in 'list' has
     additional properties, these will not effect the result. */

bool isInList(SpeciesList *list, SpeciesItem *item, bool ignoreName) {
    SpeciesListItem *currentItem;
    SpeciesItem *current;
    PropertyItem *aProp, *bProp;
    bool foundPropertyMatch = true;
    
    currentItem = list->first;

    while (currentItem) {
        current = currentItem->target;

        foundPropertyMatch = true; // To do with the scenario that there are no properties to compare
        
        if (ignoreName || (caseInsensitiveStringCompare(current->name, item->name) == 0)) {
            if ((current->living == item->living) && (current->form == item->form) && (current->type == item->type)) {
                aProp = current->firstUnknown;

                while (aProp) {
                    foundPropertyMatch = false;

                    bProp = item->firstUnknown;

                    while (bProp) {
                        if (caseInsensitiveStringCompare(aProp->name, bProp->name) == 0)  {
                            foundPropertyMatch = true;
                            break;
                        }

                        bProp = bProp->next;
                    }

                    if (!foundPropertyMatch) {
                        break;
                    }

                    aProp = aProp->next;
                }

                if (foundPropertyMatch) {
                    aProp = current->firstProperty;

                    while (aProp) {
                        foundPropertyMatch = false;

                        bProp = item->firstProperty;

                        while (bProp) {
                            if ((caseInsensitiveStringCompare(aProp->name, bProp->name) == 0) && (caseInsensitiveStringCompare(aProp->property, bProp->property) == 0)) {
                                foundPropertyMatch = true;
                                break;
                            }

                            bProp = bProp->next;
                        }

                        if (!foundPropertyMatch) {
                            break;
                        }

                        aProp = aProp->next;
                    }
                }

                if (foundPropertyMatch) {
                    return true;
                }
            }
        }
        
        currentItem = currentItem->next;
    }

    return false;
}

/* exactIsInList(SpeciesList *list, SpeciesItem *item)
     Searches 'list' for the exact instance of 'item'.  That is, it searches not
     based on contents but for that exact location in memory. */

bool exactIsInList(SpeciesList *list, SpeciesItem *item) {
    SpeciesListItem *currentItem;

    currentItem = list->first;

    while (currentItem) {
        if (currentItem->target == item) {
            return true;
        }

        currentItem = currentItem->next;
    }

    return false;
}

/* questionIsRelevant(SpeciesList *list, QuestionItem *item)
     Returns true if the question of 'item' is relevant to 'list'.  This is in
     terms of whether asking it makes any sense for that list - it is *not*
     based necessarily on whether the properties of 'item' appear in 'list', or
     similar.  e.g. if all items in 'list' have the property of 'item', this
     will return false - asking that question won't help you gain any further
     info on what species you are looking for. */

bool questionIsRelevant(SpeciesList *list, QuestionItem *item) {
    SpeciesListItem *currentItem = NULL;
    SpeciesItem *current = NULL;
    PropertyItem *temp = NULL;
    char **values = NULL;
    int itemCount = -1, count = 0, i, j, sizeofValues = 0;
    bool haveHadTrue = false, haveHadFalse = false, foundMatch = false, globalFoundMatch = false;
    
    if (!list || !item || (list->numberOfItems < 1)) {
        return false; // one could argue :)
    }

    currentItem = list->first;

    if (!(item->isBoolean)) {
        sizeofValues = list->numberOfItems;
        values = new char*[sizeofValues];
        nullifyStringArray(values, sizeofValues);
    }

    while (currentItem) {
        ++itemCount;
        current = currentItem->target;

        foundMatch = false;
        count = numberOfProperties(current);

        for (i = 0; i < count; ++i) {
            temp = propertyByIndex(current, i);

            if (temp) {
                if (caseInsensitiveStringCompare(temp->name, item->question) == 0) {
                    globalFoundMatch = true;
                    foundMatch = true;
                    
                    if (item->isBoolean) {
                        if ((temp->property == NULL) || (boolFromString(temp->property) == true)) {
                            if (haveHadFalse) {
                                deleteProperty(temp);
                                return true;
                            } else {
                                haveHadTrue = true;
                            }
                        } else {
                            if (haveHadTrue) {
                                deleteProperty(temp);
                                return true;
                            } else {
                                haveHadFalse = true;
                            }
                        }
                    } else {
                        values[itemCount] = copyOfString(temp->property);
                        
                        for (j = 0; j < itemCount; ++j) {
                            if ((values[j] != NULL) && (values[itemCount] != NULL) && (caseInsensitiveStringCompare(values[j], values[itemCount]) != 0)) {
                                deleteProperty(temp);
                                deleteStringArray(values, sizeofValues);
                                delete[] values;
                                return true;
                            } 
                        }
                    }
                }

                deleteProperty(temp);
            }
        }

        if (!foundMatch) {
            if (item->isBoolean) {
                if (haveHadTrue) {
                    return true; // no property (meaning false), but have had true previously -> relevant
                } else {
                    haveHadFalse = true;
                }
            }
        }

        currentItem = currentItem->next;
    }

    if (!(item->isBoolean)) {
        deleteStringArray(values, sizeofValues);
        delete[] values;
    }

    return false;
}

/* removeSpeciesWithProperty(SpeciesList *list, const char *name)
     Removes all species which have the property 'name', from 'list'.  Note that
     only boolean properties are matched (those where their value == NULL). */

void removeSpeciesWithProperty(SpeciesList *list, const char *name) {
    SpeciesListItem *currentItem = NULL, *nextItem = NULL;
    SpeciesItem *current = NULL;
    PropertyItem *curProp = NULL;
    int i = 0;

    if (list && name) {
        currentItem = list->first;

        while (currentItem) {
            nextItem = currentItem->next;
            current = currentItem->target;

            for (i = 0; i < numberOfProperties(current); ++i) {
                curProp = propertyByIndex(current, i);

                if ((curProp->name != NULL) && (caseInsensitiveStringCompare(curProp->name, name) == 0)) {
                    removeSpeciesFromList(currentItem);
                    deleteListItem(currentItem);
                    deleteProperty(curProp);
                    break;
                }
                
                deleteProperty(curProp);
            }

            currentItem = nextItem;
        }
    }
}

void removeSpeciesWithDifferentProperty(SpeciesList *list, const char *name, const char *value) {
    SpeciesListItem *currentItem = NULL, *tempItem = NULL;
    SpeciesItem *current = NULL;
    PropertyItem *curProp = NULL;
    bool didRemoveItem = false, foundMatch = false;
    int i;

    if (list && name) {
        currentItem = list->first;

        while (currentItem) {
            current = currentItem->target;

            didRemoveItem = false;
            foundMatch = false;
            
            for (i = 0; i < numberOfProperties(current); ++i) {
                curProp = propertyByIndex(current, i);

                if (caseInsensitiveStringCompare(curProp->name, name) == 0) {
                    foundMatch = true;
                    
                    if ((value != NULL) && (caseInsensitiveStringCompare(curProp->property, value) != 0)) {
                        tempItem = currentItem->next;
                        removeSpeciesFromList(currentItem);
                        deleteListItem(currentItem);
                        currentItem = tempItem;

                        didRemoveItem = true;
                        
                        deleteProperty(curProp);
                        break;
                    }
                }

                deleteProperty(curProp);
            }

            if (!didRemoveItem) {
                if (!foundMatch && (value == NULL)) { // boolean property, and this didn't have it
                    tempItem = currentItem->next;
                    removeSpeciesFromList(currentItem);
                    deleteListItem(currentItem);
                    currentItem = tempItem;
                } else {
                    currentItem = currentItem->next;
                }
            }
        }
    }
}

/* findByLiving(SpeciesList *list, int living)
     Searches 'list' for the first SpeciesItem with a living status equal to
     'living'.  Returns a pointer to this item, or NULL if no matches are found. */

SpeciesListItem* findByLiving(SpeciesList *list, int living) {
    SpeciesListItem *currentItem;

    currentItem = list->first;

    while (currentItem) {
        if (currentItem->target->living == living) {
            return currentItem;
        }

        currentItem = currentItem->next;
    }

    return NULL;
}

/* findByForm(SpeciesList *list, int form)
     Searches 'list' for the first SpeciesItem with a form of 'form'.  Returns a
     pointer to it's list item, or NULL if no matches are found. */

SpeciesListItem* findByForm(SpeciesList *list, int form) {
    SpeciesListItem *currentItem;

    currentItem = list->first;

    while (currentItem) {
        if (currentItem->target->form == form) {
            return currentItem;
        }

        currentItem = currentItem->next;
    }

    return NULL;
}

/* findByType(SpeciesList *list, int type)
     Searches 'list' for the first SpeciesItem with a type of 'type'.  Returns a
     pointer to it's list item, or NULL if no matches are found. */

SpeciesListItem* findByType(SpeciesList *list, int type) {
    SpeciesListItem *currentItem;

    currentItem = list->first;

    while (currentItem) {
        if (currentItem->target->type == type) {
            return currentItem;
        }

        currentItem = currentItem->next;
    }

    return NULL;
}

/* findByName(SpeciesList *list, const char *name)
     Using the same process as findSpeciesByLength, instead compares the name of
     each SpeciesItem with the name provided, returning the first matching result,
     or NULL if none. */

SpeciesListItem* findByName(SpeciesList *list, const char *name) {
    SpeciesListItem *currentItem;

    currentItem = list->first;

    while (currentItem) {
        if (!caseInsensitiveStringCompare(currentItem->target->name, name)) {
            return currentItem;
        }

        currentItem = currentItem->next;
    }

    return NULL;
}

/* findByIndex(SpeciesList *list, int val)
     Finds nth SpeciesItem, where n is 'val', in the list, starting from the first.
     Returns a pointer to said SpeciesItem, or NULL if the index requested is
     outside the scope of the list. */

SpeciesListItem* findByIndex(SpeciesList *list, int val) {
    SpeciesListItem *currentItem;

    if (val < 0) {
        return NULL;
    }

    currentItem = list->first;

    while (currentItem && (val > 0)) {
        currentItem = currentItem->next;
        --val;
    }

    return currentItem;
}

/* findNextByLiving(int living, SpeciesListItem *start)
     Continues searching from 'start' for the next SpeciesItem with a living
     status of 'living'.  Returns a pointer to this SpeciesItem list item.
     Returns NULL if no matches are found. */

SpeciesListItem* findNextByLiving(int living, SpeciesListItem *start) {
    SpeciesListItem *currentItem;

    if (start) {
        currentItem = start->next;

        while (currentItem) {
            if (currentItem->target->living == living) {
                return currentItem;
            }

            currentItem = currentItem->next;
        }
    }
    
    return NULL;
}

/* findNextByForm(int form, SpeciesListItem *start)
     Continues searching from 'start' for the next SpeciesItem with a form of
     'form'.  Returns a pointer to this SpeciesItem list item.  Returns NULL if
     no matches are found. */

SpeciesListItem* findNextByForm(int form, SpeciesListItem *start) {
    SpeciesListItem *currentItem;

    if (start) {
        currentItem = start->next;

        while (currentItem) {
            if (currentItem->target->form == form) {
                return currentItem;
            }

            currentItem = currentItem->next;
        }
    }

    return NULL;
}

/* findNextByType(int type, SpeciesListItem *start)
     Continues searching from 'start' for the next SpeciesItem with a form of
     'form'.  Returns a pointer to this SpeciesItem list item.  Returns NULL if
     no matches are found. */

SpeciesListItem* findNextByType(int type, SpeciesListItem *start) {
    SpeciesListItem *currentItem;

    if (start) {
        currentItem = start->next;

        while (currentItem) {
            if (currentItem->target->type == type) {
                return currentItem;
            }

            currentItem = currentItem->next;
        }
    }

    return NULL;
}

/* findNextByName(const char *name, SpeciesListItem *start)
     Using the same process as findNextSpeciesByLength, instead compares the title
     of each SpeciesItem with the name provided, returning the next matching result
     after 'start', or NULL if none. */

SpeciesListItem* findNextByName(const char *name, SpeciesListItem *start) {
    SpeciesListItem *currentItem;

    if (start) {
        currentItem = start->next;

        while (currentItem) {
            if (!caseInsensitiveStringCompare(currentItem->target->name, name)) {
                return currentItem;
            }

            currentItem = currentItem->next;
        }
    }
    
    return NULL;
}

/* findAllByLiving(SpeciesList *list, int living)
     Returns a SpeciesList instance containing all results for the specified
     search in 'list'. */

SpeciesList* findAllByLiving(SpeciesList *list, int living) {
    SpeciesList *result;
    SpeciesListItem *currentItem;

    result = newList(NULL);
    currentItem = findByLiving(list, living);

    while (currentItem) {
        copySpeciesToList(result, currentItem);
        currentItem = findNextByLiving(living, currentItem);
    }

    return result;
}

/* findAllByForm(SpeciesList *list, int form)
     Returns a SpeciesList instance containing all results for the specified
     search in 'list'. */

SpeciesList* findAllByForm(SpeciesList *list, int form) {
    SpeciesList *result;
    SpeciesListItem *currentItem;

    result = newList(NULL);
    currentItem = findByForm(list, form);

    while (currentItem) {
        copySpeciesToList(result, currentItem);
        currentItem = findNextByForm(form, currentItem);
    }

    return result;
}

/* findAllByType(SpeciesList *list, int type)
     Returns a SpeciesList instance containing all results for the specified
     search in 'list'. */

SpeciesList* findAllByType(SpeciesList *list, int type) {
    SpeciesList *result;
    SpeciesListItem *currentItem;

    result = newList(NULL);
    currentItem = findByType(list, type);

    while (currentItem) {
        copySpeciesToList(result, currentItem);
        currentItem = findNextByType(type, currentItem);
    }

    return result;
}

/* findAllByName(SpeciesList *list, const char *name)
     Returns a SpeciesList instance containing all results for the specified
     search in the list 'list'.  Uses the functions findSpeciesByTitle() and
     findNextSpeciesByTitle(), so the 'name' parameter functions the same. */

SpeciesList* findAllByName(SpeciesList *list, const char *name) {
    SpeciesList *result;
    SpeciesListItem *currentItem;

    result = newList(NULL);
    currentItem = findByName(list, name);

    while (currentItem) {
        copySpeciesToList(result, currentItem);
        currentItem = findNextByName(name, currentItem);
    }

    return result;
}

/* saveListToFile(SpeciesList *list, const char *fileName)
     If 'filename' is not NULL, writes the contents of 'list' to it.  If it is
     NULL, the 'openedFrom' property of the 'list' will be used. Returns true if
     the save was successful, false otherwise.  Pass NULL as the 'fileName'
     parameter to use the openedFrom property of 'list'.
       If 'strict' is true, saves only data supported by the basic file format,
     prompting the user if a situation arises in which data will be lost, and
     giving the user the option to cancel the save, save the extra data, or lose
     the extra data. */

bool saveListToFile(SpeciesList *list, const char *fileName, bool strict) {
    ofstream fout;
    SpeciesListItem *currentItem;
    SpeciesItem *current;
    PropertyItem *curProp;
    char *holder;
    int propCount;
    bool didSkipData = false, propLimitOverride = false;
    
    if (fileName) {
        fout.open(fileName);
    } else {
        fout.open(list->openedFrom);
    }

    if (fout.fail()) {
        return false; // unable to open file
    }

    if (strict) {
        cout << "Saving in strict format...\n";
    } else {
        cout << "Saving...\n";
    }
    
    currentItem = list->first;
    
    while (currentItem) {
        current = currentItem->target;
        propCount = 0;
        propLimitOverride = false;
        
        fout << current->name << endl;

        if (current->living != UNKNOWN) {
            if (strict && (current->living == LIVING_NO)) {
                cout << "The species \"" << current->name << "\" is marked as being extinct, but this cannot be saved\nimplicitly in strict mode.  Override and save anyway?  ";

                if (getBoolean(cin)) {
                    cout << "Saving anyway.\n";
                    ++propCount;
                    holder = livingToString(current->living, true);

                    if (holder) {
                        fout << holder << endl;
                        delete[] holder;
                    }
                } else {
                    didSkipData = true;
                    cout << "Skipping.\n";
                }
            } else {
                ++propCount;
                holder = livingToString(current->living, true);

                if (holder) {
                    fout << holder << endl;
                    delete[] holder;
                }
            }
        }

        if (current->form != UNKNOWN) {
            ++propCount;
            holder = formToString(current->form, true);

            if (holder) {
                fout << holder << endl;
                delete[] holder;
            }
        }

        if (current->type != UNKNOWN) {
            if (strict && (current->type == ANIMAL_AMPHIBIAN)) {
                cout << "The species \"" << current->name << "\" is listed as an Amphibian, but this type is not\nsupported in strict format mode.  Override and save anyway?  ";

                if (getBoolean(cin)) {
                    cout << "Saving anyway.\n";
                    ++propCount;
                    holder = typeToString(current->type, true);

                    if (holder) {
                        fout << holder << endl;
                        delete[] holder;
                    }
                } else {
                    didSkipData = true;
                    cout << "Skipping.\n";
                }
            } else {
                ++propCount;
                holder = typeToString(current->type, true);

                if (holder) {
                    fout << holder << endl;
                    delete[] holder;
                }
            }
        }

        curProp = current->firstUnknown;

        while (curProp) {
            ++propCount;

            if (strict && (propCount == STRICT_PROPERTY_COUNT_LIMIT + 1)) {
                cout << "Strict mode allows for up to only " << STRICT_PROPERTY_COUNT_LIMIT << " properties per species.\nThe species \"" << current->name << "\" has more.   Override and save anyway?  ";

                if (getBoolean(cin)) {
                    cout << "Saving anyway.\n";
                    propLimitOverride = true;
                    fout << curProp->name << endl;
                } else {
                    didSkipData = true;
                    cout << "Skipping.\n";
                }
            } else if (!strict || (propCount <= STRICT_PROPERTY_COUNT_LIMIT) || propLimitOverride) {
                fout << curProp->name << endl;
            }

            curProp = curProp->next;
        }
        
        curProp = current->firstProperty;

        while (curProp) {
            ++propCount;
            
            if (strict && (propCount == STRICT_PROPERTY_COUNT_LIMIT + 1)) {
                cout << "Strict mode allows for up to only " << STRICT_PROPERTY_COUNT_LIMIT << " properties per species.\nThe species \"" << current->name << "\" has more.   Override and save anyway?  ";

                if (getBoolean(cin)) {
                    cout << "Saving anyway.\n";
                    propLimitOverride = true;
                    fout << '#' << curProp->name << ':' << curProp->property << endl;
                } else {
                    didSkipData = true;
                    cout << "Skipping.\n";
                }
            } else if (!strict || (propCount <= STRICT_PROPERTY_COUNT_LIMIT) || propLimitOverride) {
                fout << '#' << curProp->name << ':' << curProp->property << endl;
            }

            curProp = curProp->next;
        }
        
        currentItem = currentItem->next;

        if (currentItem) {
            fout << endl;
        }
    }

    fout.close();

    if (fileName) {
        if (didSkipData) {
            cout << "Successfully saved some data from list \"" << list->openedFrom << "\" to file \"" << fileName << "\".\n";
        } else {
            if (list->numberOfItems == 0) {
                cout << "Successfully saved empty list to \"" << fileName << "\".\n";
            } else {
                cout << "Successfully saved all data in list \"" << list->openedFrom << "\" to file \"" << fileName << "\".\n";
            }
        }
    } else {
        if (didSkipData) {
            cout << "Successfully saved some data from list \"" << list->openedFrom << "\" to file.\n";
        } else {
            if (list->numberOfItems == 0) {
                cout << "Successfully saved empty list to \"" << list->openedFrom << "\".\n";
            } else {
                cout << "Successfully saved all data in list \"" << list->openedFrom << "\" to file.\n";
            }
        }
    }
    
    return true;
}

/* saveListAs(SpeciesList *list, bool strict)
     Prompts the user for a file name and saves the database 'list' to it.  Returns
     false if saving was unsuccessful *and* the user choose not to quit anyway.
     If saving failed but they chose to quit, returns true.
       If 'strict' is true, strict file formatting is enforced, otherwise
     extended file formatting is allowed. */

bool saveListAs(SpeciesList *list, bool strict) { // Marked not finished, but why?  Watch for bugs...
    char input[FILENAME_BUFFER_SIZE];

    if (list == NULL) {
        return true;
    }

    cout << "Do you wish to save the list to file?  ";

    if (!getBoolean(cin)) {
        return true;
    }
                   
    if (list->openedFrom == NULL) {
        cout << "Please enter the file name to save the database to (enter nothing to not save):  ";
    } else {
        cout << "Please enter the file name to save the database to (press return for \"" << list->openedFrom << "\"):  ";
    }
    
    safeGetline(cin, input, FILENAME_BUFFER_SIZE);

    if (strlen(input)) {
        if (saveListToFile(list, input, strict)) {
            return true;
        } else {
            cout << "\nUnable to save changes - the file \"" << input << "\" may be unwritable or already in use.\nDo you wish to quit without saving changes?  ";

            if (getBoolean(cin)) {
                cout << "\nQuitting without saving changes.  Goodbye.\n\n";
                return true;
            } else {
                cout << "\nCancelling quit - returning to main menu.";
                return false;
            }
        }
    } else {
        if (list->openedFrom != NULL) {
            return (saveListToFile(list, list->openedFrom, strict));
        } else {
            cout << "\nQuitting without saving.  Goodbye.\n\n";
            return true;
        }
    }
}

/* deriveFormFromType(int form)
     Uses the form to determine what type the species is.  Useful when the form
     is missing for some reason, and for checking it is valid if it is known. */

int deriveFormFromType(int form) {
    switch (form) {
    case ANIMAL_LAND:
    case ANIMAL_WATER:
    case ANIMAL_AIR:
    case ANIMAL_AMPHIBIAN:
        return ANIMAL;
    case PLANT_TREE:
    case PLANT_FLOWER:
    case PLANT_SHRUB:
        return PLANT;
    default:
        return UNKNOWN;
    }
}

/* livingFromString(const char *data)
     Returns the appropriate type (int) for the living status specified in the
     string 'data'.  Not case sensitive.  Returns unknown if unable to match. */

int livingFromString(const char *data) {
    if (data) {
        if (caseInsensitiveStringCompare(data, "Living") == 0) {
            return LIVING_YES;
        } else if (caseInsensitiveStringCompare(data, "Extinct") == 0) {
            return LIVING_NO;
        } else {
            return UNKNOWN;
        }
    } else {
        return UNKNOWN;
    }
}

/* livingToString(int form, bool allowUnknown)
     Returns a string representing the user-readable name for the living status
     given.  If 'allowUnknown' is true, "Unknown" will be returned in the case
     that no match can be made.  If it is false, NULL will be returned instead. */

char* livingToString(int living, bool allowUnknown) {
    char* result = NULL;

    switch (living) {
    case LIVING_NO:
        result = new char[8];
        strncpy(result, "Extinct", 8);
        break;
    case LIVING_YES:
        result = new char[7];
        strncpy(result, "Living", 7);
        break;
    default:
        if (allowUnknown) {
            result = new char[8];
            strncpy(result, "Unknown", 8);
        }
    }

    return result;
}

/* formFromString(const char *data)
     Returns the appropriate type (int) for the form specified in the string
     'data'.  Not case sensitive.  Returns unknown if unable to match. */

int formFromString(const char *data) {
    if (data) {
        if (caseInsensitiveStringCompare(data, "Animal") == 0) {
            return ANIMAL;
        } else if (caseInsensitiveStringCompare(data, "Plant") == 0) {
            return PLANT;
        } else {
            return UNKNOWN;
        }
    } else {
        return UNKNOWN;
    }
}

/* formToString(int form, bool allowUnknown)
     Returns a string representing the user-readable name for the form given.
     If 'allowUnknown' is true, "Unknown" will be returned in the case that no
     match can be made.  If it is false, NULL will be returned instead.*/

char* formToString(int form, bool allowUnknown) {
    char* result = NULL;

    switch (form) {
    case ANIMAL:
        result = new char[7];
        strncpy(result, "Animal", 7);
        break;
    case PLANT:
        result = new char[6];
        strncpy(result, "Plant", 6);
        break;
    default:
        if (allowUnknown) {
            result = new char[8];
            strncpy(result, "Unknown", 8);
        }
    }

    return result;
}

/* typeFromString(const char *data)
     Returns the appropriate type (int) for the type specified in the string 'data'.
     Not case sensitive.  Returns unknown if unable to match. */

int typeFromString(const char *data) {
    if (data) {
        if (caseInsensitiveStringCompare(data, "Land Animal") == 0) {
            return ANIMAL_LAND;
        } else if (caseInsensitiveStringCompare(data, "Water Animal") == 0) {
            return ANIMAL_WATER;
        } else if (caseInsensitiveStringCompare(data, "Air Animal") == 0) {
            return ANIMAL_AIR;
        } else if (caseInsensitiveStringCompare(data, "Amphibian") == 0) {
            return ANIMAL_AMPHIBIAN;
        } else if (caseInsensitiveStringCompare(data, "Tree") == 0) {
            return PLANT_TREE;
        } else if (caseInsensitiveStringCompare(data, "Flower") == 0) {
            return PLANT_FLOWER;
        } else if (caseInsensitiveStringCompare(data, "Shrub") == 0) {
            return PLANT_SHRUB;
        } else {
            return UNKNOWN;
        }
    } else {
        return UNKNOWN;
    }
}

/* typeToString(int type, bool allowUnknown)
     Returns a string representing the user-readable name for the type given.  If
     'allowUnknown' is true, "Unknown" will be returned in the case that no match
     can be made.  If it is false, NULL will be returned instead.*/

char* typeToString(int type, bool allowUnknown) {
    char* result = NULL;

    switch (type) {
    case ANIMAL_LAND:
        result = new char[12];
        strncpy(result, "Land Animal", 12);
        break;
    case ANIMAL_WATER:
        result = new char[13];
        strncpy(result, "Water Animal", 13);
        break;
    case ANIMAL_AIR:
        result = new char[11];
        strncpy(result, "Air Animal", 11);
        break;
    case ANIMAL_AMPHIBIAN:
        result = new char[10];
        strncpy(result, "Amphibian", 10);
        break;
    case PLANT_TREE:
        result = new char[5];
        strncpy(result, "Tree", 5);
        break;
    case PLANT_FLOWER:
        result = new char[7];
        strncpy(result, "Flower", 7);
        break;
    case PLANT_SHRUB:
        result = new char[6];
        strncpy(result, "Shrub", 6);
        break;
    default:
        if (allowUnknown) {
            result = new char[8];
            strncpy(result, "Unknown", 8);
        }
    }

    return result;
}

/* determinePropertyFromString(const char *data, int &value)
     Determines what type of property (living, form or type) the passed 'data' is
     and returns that.  Sets 'value' to the appropriate value for that type. */

int determinePropertyFromString(const char *data, int &value) {
    value = UNKNOWN;
    
    if (data && strlen(data)) {
        switch (toupper(data[0])) {
        case 'A':
            if (caseInsensitiveStringCompare(data, "Air Animal") == 0) {
                value = ANIMAL_AIR;
                return PROP_TYPE;
            } else if (caseInsensitiveStringCompare(data, "Animal") == 0) {
                value = ANIMAL;
                return PROP_FORM;
            } else if (caseInsensitiveStringCompare(data, "Amphibian") == 0) {
                value = ANIMAL_AMPHIBIAN;
                return PROP_TYPE;
            } else {
                return UNKNOWN;
            }
        case 'P':
            if (caseInsensitiveStringCompare(data, "Plant") == 0) {
                value = PLANT;
                return PROP_FORM;
            } else {
                return UNKNOWN;
            }
        case 'L':
            if (caseInsensitiveStringCompare(data, "Land Animal") == 0) {
                value = ANIMAL_LAND;
                return PROP_TYPE;
            } else if (caseInsensitiveStringCompare(data, "Living") == 0) {
                value = LIVING_YES;
                return PROP_LIVING;
            } else {
                return UNKNOWN;
            }
        case 'W':
            if (caseInsensitiveStringCompare(data, "Water Animal") == 0) {
                value = ANIMAL_WATER;
                return PROP_TYPE;
            } else {
                return UNKNOWN;
            }
        case 'T':
            if (caseInsensitiveStringCompare(data, "Tree") == 0) {
                value = PLANT_TREE;
                return PROP_TYPE;
            } else {
                return UNKNOWN;
            }
        case 'F':
            if (caseInsensitiveStringCompare(data, "Flower") == 0) {
                value = PLANT_FLOWER;
                return PROP_TYPE;
            } else {
                return UNKNOWN;
            }
        case 'S':
            if (caseInsensitiveStringCompare(data, "Shrub") == 0) {
                value = PLANT_SHRUB;
                return PROP_TYPE;
            } else {
                return UNKNOWN;
            }
        case 'E':
            if (caseInsensitiveStringCompare(data, "Extinct") == 0) {
                value = LIVING_NO;
                return PROP_LIVING;
            } else {
                return UNKNOWN;
            }
        default:
            return UNKNOWN;
        }
    } else {
        return UNKNOWN;
    }
}

/* copyOfString(const char *data)
     Returns a copy of the string 'data'. */

char* copyOfString(const char *data) {
    char *result;
    int length;

    length = strlen(data) + 1;

    result = new char[length];
    strncpy(result, data, length);

    return result;
}

/* boolToString(bool value)
     Converts the boolean passed to a string value.  Defaults to true = "Yes",
     false = "No". */

char* boolToString(bool value) {
    char *result;

    if (value) {
        result = new char[4];
        strncpy(result, "Yes", 4);
    } else {
        result = new char[3];
        strncpy(result, "No", 3);
    }

    return result;
}

/* boolFromString(const char *data)
     Converts the string passed to a boolean value, using a wide variety of
     techniques to hopefully maximize compatability.  Returns false by default. */

bool boolFromString(const char *data) {
    if (strlen(data)) {
        switch (toupper(data[0])) {
        case 'Y': // Yes
            return true;
        case 'N': // No/Negative
            return false;
        case '1': // 1
            return true;
        case '0': // 0
            return false;
        case 'T': // True
            return true;
        case 'F': // False
            return false;
        case '+': // +
            return true;
        case '-': // -
            return false;
        }
    }

    return false;
}

/* isBoolean(const char *data)
     Returns whether or not the passed string contains a boolean value [so far as
     it can tell]. */

bool isBoolean(const char *data) {
    char smallTemp[2] = {'\0', '\0'};

    if (data) {
        if (strlen(data) == 1) {
            smallTemp[0] = toupper(data[0]);

            return (strpbrk(smallTemp, "YNTF10+-") != NULL);
        } else {
            return (caseInsensitiveStringCompare(data, "yes") == 0 ||
                    caseInsensitiveStringCompare(data, "no") == 0 ||
                    caseInsensitiveStringCompare(data, "true") == 0 ||
                    caseInsensitiveStringCompare(data, "false") == 0);
        }
    } else {
        return false;
    }
}

/* caseInsensitiveStringCompare(const char *a, const char *b)
     Compares the two strings 'a' and 'b' without regard to their case.  Returns
     -1 if 'a' is less than 'b', 1 if 'b' is less than 'a', and 0 if they are
     identical - i.e. the same as strcmp(). */

int caseInsensitiveStringCompare(const char *a, const char *b) {
    int offset, end, lengthA, lengthB, charA, charB;

    lengthA = strlen(a);
    lengthB = strlen(b);
    
    end = (lengthA < lengthB) ? lengthA : lengthB;

    for(offset = 0; offset < end; ++offset) {
        charA = toupper(a[offset]);
        charB = toupper(b[offset]);

        if (charA != charB) {
            return (charA < charB) ? -1 : 1;
        }
    }

    if (lengthA == lengthB) {
        return 0;
    } else {
        return (lengthA < lengthB) ? -1 : 1;
    }
}

/* safeGetline(istream &in, char *data, int maxLength)
     Same general functionality as getline in terms of input retrieval, except
     for two things.  First, in the case where there is too much data it clears
     the error flag and buffer [up to the next line] so "in" may be used again.
       Second, it eliminates all whitespace at the start and end of the input
     string. */

void safeGetline(istream& in, char *data, int maxLength) {
    char *temp;
    
    if (in.fail()) { // Stream already broken
        return;
    }

    temp = new char[maxLength];
    in.getline(temp, maxLength - 1);

    if (in.fail() && strlen(temp)) {
        in.clear();
        in.ignore(10000, '\n');
    }

    removeOutsideWhitespace(&temp);
    strncpy(data, temp, strlen(temp) + 1);
    delete[] temp;
}

/* getBoolean(istream& in)
     Retrieves a yes or no answer from the user.  Doesn't display any prompt
     initially, but will advise the user if they enter something invalid. */

bool getBoolean(istream& in) {
    char temp[LINE_BUFFER_SIZE];

    if (in.fail()) { // Stream broken
        return false;
    }

    while (1) {
        safeGetline(cin, temp, LINE_BUFFER_SIZE);

        if (isBoolean(temp)) {
            return boolFromString(temp);
        } else {
            cout << "Please enter a yes or no answer, e.g. 'yes' or 'no':  ";
        }
    }
}

/* getChar(istream& in)
     Returns a character input from 'in', and clears the input stream until the
     next new line. */

char getChar(istream& in) {
    char temp;

    if (in.fail()) { // Stream broken
        return '\0';
    }

    in >> temp;
    in.ignore(10000, '\n');

    return temp;
}

/* switchIsInArguments(int argc, const char **argv, char sw)
     Returns true if the switch 'sw' appears anywhere in the standard argument
     list 'argv'.  Allows for any number of switches per field, e.g. both "-p"
     and "-cFp" will return true for 'p'.  Case sensitive. */

bool switchIsInArguments(int argc, const char **argv, char sw) {
    for (int i = 1; i < argc; ++i) {
        if (argv[i][0] == '-') { // switch
            for (unsigned int n = 1; n < strlen(argv[i]); ++n) {
                if (argv[i][n] == sw) {
                    return true;
                }
            }
        }
    }

    return false;
}

/* firstDataArgument(int argc, const char **argv)
     Returns the index of the first data argument - i.e. the first argument which
     doesn't start with a '-'.  In the event that none is found, returns 0. */

int firstDataArgument(int argc, const char **argv) {
    for (int i = 1; i < argc; ++i) {
        if (argv[i][0] != '-') { // switch
            return i;
        }
    }

    return 0;
}

/* isArrayVoid(const void *target[], int n)
     Checks the pointer array 'target', containing 'n' items, to see if it is
     completely NULL.  Returns true in this case, false otherwise. */

bool isArrayVoid(void *target[], int n) {
    for (int i = 0; i < n; ++i) {
        if (target[i] != NULL) {
            return false;
        }
    }

    return true;
}

/* deleteStringArray(char *target[], int n)
     Deletes 'n' strings from the array 'target'. */

void deleteStringArray(char *target[], int n) {
    for (int i = 0; i < n; ++i) {
        if (target[i] != NULL) {
            delete[] target[i];
            target[i] = NULL;
        }
    }
}

/* nullifyStringArray(char *target[], int n)
     Sets every entry in 'target', where the number of entries is 'n', to NULL. */

void nullifyStringArray(char *target[], int n) {
    for (int i = 0; i < n; ++i) {
        target[i] = NULL;
    }
}

/* propertyNameFromLine(const char *data)
     Returns a new cstring containing just the name part of a property line.  This
     removes the initial # and terminates one character short of the colon, i.e.
     using it on "#Name:yadda" would yield "Name". */

char* propertyNameFromLine(const char *data) {
    char *result = NULL;
    int length = 1, i = 1;

    while(data[i] != ':') {
        ++i;
        ++length;
    }

    result = new char[length];

    for(i = 0; i < length; ++i) {
        result[i] = data[i + 1];
    }

    result[length - 1] = '\0';
    
    return result;
}

/* propertyValueFromLine(const char *data)
     Returns a new cstring containing just the value part of the property line
     'data'. */

char* propertyValueFromLine(const char *data) {
    char *result = NULL;
    int start = 2, length = 1, i = 1;

    while(data[i] != ':') {
        ++i;
        ++start;
    }

    i = start;

    while(data[i] != '\0') {
        ++i;
        ++length;
    }

    result = new char[length];

    for(i = 0; i < length; ++i) {
        result[i] = data[i + start];
    }

    return result;
}
