/*  Student Name: Wade Tregaskis
    Student Number: 02557793
    Estimated time spent on this Stage 1 assignment: Unknown at time of writing

    Notes:  Project created by creating a copy of the IPC assignment, and using
    that as a starting point.  The reason for this is that said assignment already
    contains lots of code for linked lists, save/load of data etc, as well as many
    utility functions (e.g. caseInsensitiveStringCompare). */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef SPECIES_H
#define SPECIES_H

// -INCLUDES-

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cctype>
#include <cstring>
#include <cmath>
#include <stdarg.h>
//#include <mcheck.h> // for generating trace logs
using namespace std;


// -DEFINES-

#define SPECIES_H_VERSION 1.04
#define STRICT_PROPERTY_COUNT_LIMIT 10
#define BRIEF_LIST_THRESHOLD 4
#define LINE_BUFFER_SIZE 251
#define FILENAME_BUFFER_SIZE 255
#define UNKNOWN -1


// -ENUMS-

/* Note the use of increasing orders of magnitude in the enums below.
   This isn't important at time of writing, but leaves room for future use.
     Also note that no enum type will EVER be 0.  All valid values will be
   positive, while 'errors' will be negative (obviously != UNKNOWN). */

enum { LIVING_NO = 1, LIVING_YES = 2 };
enum { ANIMAL = 10, PLANT = 20 };
enum { ANIMAL_LAND = 100, ANIMAL_WATER = 200, ANIMAL_AIR = 300, ANIMAL_AMPHIBIAN = 400, PLANT_TREE = 500, PLANT_FLOWER = 600, PLANT_SHRUB = 700 };
enum { PROP_LIVING = 10000, PROP_FORM = 20000, PROP_TYPE = 30000 };


// -STRUCTS-

struct SpeciesItem;
struct SpeciesListItem;
struct SpeciesList;
struct PropertyItem;
struct QuestionItem;

// The sizes in the comments were used for memory debugging, if you must know

struct SpeciesItem { // 32 bytes
    char *name;
    int living, form, type, dataError, claims;
    PropertyItem *firstProperty, *firstUnknown;
};

struct SpeciesListItem { // 20 bytes
    SpeciesListItem *previous, *next, *original;
    SpeciesList *owner;
    SpeciesItem *target;
};

struct SpeciesList { // 20 bytes
    SpeciesListItem *first, *last;
    int numberOfItems, numberOfDataErrors;
    char *openedFrom;
};

struct PropertyItem { // 12 bytes
    char *name, *property;
    PropertyItem *next;
};

struct QuestionItem { // 16 bytes [bool = 4?]
    char *question, *answer;
    bool isBoolean;
    QuestionItem* next;
};


// -FUNCTIONS-

void uiDisplayList(SpeciesList *list);
void uiDisplayList(SpeciesList *list, bool format);
void uiAddSpeciesToList(SpeciesList *list);
void uiSearchList(SpeciesList *list);
void uiExplainDataErrors(SpeciesList *list);
void uiExplainStrictFileFormat(void);

void uiSearchList_askQuestions(SpeciesList *list, QuestionItem *startList);
void uiSearchList_showUpdate(SpeciesList *list);
bool uiSearchList_askIfCorrect(void);
bool uiSearchList_canMatch(SpeciesList *original, SpeciesList *sublist, int living, int form, int type);
bool uiSearchList_canMatch(SpeciesList *original, SpeciesList *sublist, QuestionItem *questions);
void uiSearchList_askToAdd(SpeciesList *list, int living, int form, int type);
void uiSearchList_askToAdd(SpeciesList *list, QuestionItem *questions);

int uiIndexedMenu(const char *title, int optionCount, ...);
void displayRepeat(ostream& out, char ch, int times);
void removeOutsideWhitespace(char **data);
void outputFixedLength(ostream& out, const char *data, int maxLength, bool left);
void outputItem(ostream& out, SpeciesItem *item, bool format);

void deleteList(SpeciesList *target);
void deleteListItem(SpeciesListItem *item);
void deletePropertyList(PropertyItem *list);
void deleteProperty(PropertyItem *item);
void deleteQuestionList(QuestionItem *list);
void deleteQuestion(QuestionItem *item);
void deleteSpecies(SpeciesItem *target);

SpeciesList* newList(char *sourceFile);
SpeciesList* newListFromFile(char *fileName, bool strict);
SpeciesItem* newSpecies(char *name);
SpeciesItem* newSpecies(char *name, char *form, char *type, int living);
SpeciesItem* newSpecies(char *name, int form, int type, int living, PropertyItem* firstProp);
PropertyItem* newProperty(char *name, char *value);
QuestionItem* newQuestionItem(char *question);
QuestionItem* newQuestionItem(char *question, char *answer);
QuestionItem* newQuestionItem(char *question, char *answer, bool boolean);
QuestionItem* newQuestionList(int living, int form, int type, bool ignoreUnknown);

int numberOfProperties(SpeciesItem *item);
PropertyItem* propertyByIndex(SpeciesItem *item, int n);

QuestionItem* generateNextQuestion(SpeciesList *list, QuestionItem *pastQuestions, bool tryToBeRandom);
bool questionIsInList(const char *question, QuestionItem *list);
bool isInList(SpeciesList *list, SpeciesItem *item, bool ignoreName);
bool exactIsInList(SpeciesList *list, SpeciesItem *item);
bool questionIsRelevant(SpeciesList *list, QuestionItem *item);

void removeSpeciesWithProperty(SpeciesList *list, const char *name);
void removeSpeciesWithDifferentProperty(SpeciesList *list, const char *name, const char *value);

SpeciesListItem* findByIndex(SpeciesList *list, int val);
SpeciesListItem* findByLiving(SpeciesList *list, int living);
SpeciesListItem* findByForm(SpeciesList *list, int form);
SpeciesListItem* findByType(SpeciesList *list, int type);
SpeciesListItem* findByName(SpeciesList *list, const char *name);

SpeciesListItem* findNextByLiving(int living, SpeciesListItem *start);
SpeciesListItem* findNextByForm(int form, SpeciesListItem *start);
SpeciesListItem* findNextByType(int type, SpeciesListItem *start);
SpeciesListItem* findNextByName(const char *name, SpeciesListItem *start);

SpeciesList* findAllByLiving(SpeciesList *list, int living);
SpeciesList* findAllByForm(SpeciesList *list, int form);
SpeciesList* findAllByType(SpeciesList *list, int type);
SpeciesList* findAllByName(SpeciesList *list, const char *name);

SpeciesList* copyOfList(SpeciesList *list);
void addSpeciesToList(SpeciesList *list, SpeciesItem *entry);
void copySpeciesToList(SpeciesList *list, SpeciesListItem *entry);
bool removeSpeciesFromList(SpeciesListItem *entry);
bool saveListToFile(SpeciesList *list, const char *fileName, bool strict);
bool saveListAs(SpeciesList *list, bool strict);

int deriveFormFromType(int form);
int livingFromString(const char *data);
char* livingToString(int living, bool allowUnknown);
int formFromString(const char *data);
char* formToString(int form, bool allowUnknown);
int typeFromString(const char *data);
char* typeToString(int form, bool allowUnknown);
bool boolFromString(const char *data);
char* boolToString(bool value);
bool isBoolean(const char *data);
char* propertyNameFromLine(const char *data);
char* propertyValueFromLine(const char *data);
int determinePropertyFromString(const char *data, int &value);

char* copyOfString(const char *data);
int caseInsensitiveStringCompare(const char *a, const char *b);
void safeGetline(istream& in, char *data, int maxLength);
char getChar(istream& in);
bool getBoolean(istream& in);
bool isArrayVoid(void *target[], int n);
void nullifyStringArray(char *target[], int n);
void deleteStringArray(char *target[], int n);

bool switchIsInArguments(int argc, const char **argv, char sw);
int firstDataArgument(int argc, const char **argv);

#endif
